package PEF::Front::WebSocket::Interface;
use warnings;
use strict;

sub on_message {
}

sub on_drain {
}

sub on_open {
}

sub on_error {
}

sub on_queue {
}

sub on_queue_error {
}

sub on_close {
}

1;
